#include <stdio.h>
#include "Hash.cc"
#include "STEntry.cc"
#include "SymbolTable.h"

/**
* Symbol table stores many lexeme and token pairs. This symbol table uses
* hash to speed up the serching time.
*
* @author       Taeho Oh
* @version      0.0.1
* @since        1999/03/31
*/
class SymbolTable
{
private:
	/**
	* The size of hash table.
	*/
	const int hashTableSize = 50 ;

	/**
	* The array heads of the linked list.
	*/
	STEntry *hashHead[hashTableSize] ;
public:
	/**
	* Initialize hash table to NULL.
	*/
	SymbolTable()
	{
		for(int i = 0; i < hashTableSize; i++)
			hashHead[i] = NULL ;
	}

	/**
	* Delete all allocated symbol table entry of the symbol table.
	*/
	~SymbolTable()
	{
		STEntry *deleteSTEntry ;

		for(int i = 0; i < hashTableSize; i++)
		{
			deleteSTEntry = hashHead[i] ;
			while(deleteSTEntry != NULL)
			{
				hashHead[i] = deleteSTEntry->getNext() ;
				delete deleteSTEntry ;
				deleteSTEntry = hashHead[i] ;
			}
		}
	}

	/**
	* Insert the symbol table entry which is composed of lexeme and
	* token to the symbol table.
	*
	* @param lexeme lexeme which will be stored.
	* @param token the token value of lexeme.
	* @param offset the offset address of the lexeme.
	* @param global 1 if the lexeme is global.
	* @return none.
	*/
	void insert(const char * const lexeme, const int token, int offset, int global)
	{
		int hashValue ;

		hashValue = Hash::getHashValue(lexeme,hashTableSize) ;

#ifdef _DEBUG_
		printf("\n") ;
		printf("insert to symbol table.\n") ;
		printf("lexeme : %s\n", lexeme) ;
		printf("token : %d\n", token) ;
		printf("hashValue : %d\n", hashValue) ;
#endif

		if(hashHead[hashValue] == NULL)
			hashHead[hashValue] = new STEntry(lexeme, token, offset, global) ;
		else
		{
			STEntry *newSTEntry = new STEntry(lexeme, token, offset, global) ;
			newSTEntry->setNext(hashHead[hashValue]) ;
			hashHead[hashValue] = newSTEntry ;
		}
	}

	/**
	* Remove the symbol table entry from the symbol table.
	*
	* @param lexeme lexeme which will be used to find symbol table entry.
	* @return none.
	*/
	void remove(char *lexeme)
	{
		int hashValue ;

		hashValue = Hash::getHashValue(lexeme,hashTableSize) ;

		if(hashHead[hashValue] == NULL)
		{

#ifdef _DEBUG_
			printf("\n") ;
			printf("remove the entry of symbol table.\n") ;
			printf("Can't find.\n") ;
			printf("lexeme : %s\n", lexeme) ;
			printf("hashValue : %d\n", hashValue) ;
#endif _DEBUG_

			return ;
		}
		else
		{
			STEntry *findingSTEntry ;
			findingSTEntry = hashHead[hashValue] ;

			findingSTEntry = hashHead[hashValue] ;
			if(!strcmp(findingSTEntry->getLexeme(), lexeme))
			{
				hashHead[hashValue] = findingSTEntry->getNext() ;

#ifdef _DEBUG_
				printf("\n") ;
				printf("remove the entry of symbol table\n") ;
				printf("lexeme : %s\n",findingSTEntry->getLexeme()) ;
				printf("token : %d\n",findingSTEntry->getToken()) ;
				printf("hashValue : %d\n",hashValue) ;
#endif _DEBUG_

				hashHead[hashValue] = findingSTEntry->getNext() ;
				delete findingSTEntry ;
				return ;
			}
			else
			{
				for(; findingSTEntry->getNext() != NULL; findingSTEntry = findingSTEntry->getNext())
				{
					if(!strcmp(findingSTEntry->getNext()->getLexeme(), lexeme))
					{

#ifdef _DEBUG_
						printf("\n") ;
						printf("find the entry of symbol table\n") ;
						printf("lexeme : %s\n",findingSTEntry->getLexeme()) ;
						printf("token : %d\n",findingSTEntry->getToken()) ;
						printf("hashValue : %d\n",hashValue) ;
#endif _DEBUG_

						STEntry *removeSTEntry ;
						removeSTEntry = findingSTEntry->getNext() ;
						findingSTEntry->setNext(removeSTEntry->getNext()) ;
						delete removeSTEntry ;
						return ;
					}
				}
			}
		}

#ifdef _DEBUG_
		printf("\n") ;
		printf("remove the entry of symbol table.\n") ;
		printf("Can't find.\n") ;
		printf("lexeme : %s\n", lexeme) ;
		printf("hashValue : %s\n", hashValue) ;
#endif _DEBUG_

		return ;
	}

	/**
	* Look up the symbol table entry from the symbol table.
	*
	* @param lexeme lexeme which will be used to find symbol table entry.
	* @return the token of lexeme.
	*/
	STEntry *lookup(const char * const lexeme) const
	{
		int hashValue ;

		hashValue = Hash::getHashValue(lexeme,hashTableSize) ;

		if(hashHead[hashValue] == NULL)
		{

#ifdef _DEBUG_
			printf("\n") ;
			printf("find the entry of symbol table.\n") ;
			printf("Can't find.\n") ;
			printf("lexeme : %s\n", lexeme) ;
			printf("hashValue : %d\n", hashValue) ;
#endif _DEBUG_

			return NULL ;
		}
		else
		{
			STEntry *findingSTEntry ;
			findingSTEntry = hashHead[hashValue] ;

			for(STEntry *findingSTEntry = hashHead[hashValue]; findingSTEntry != NULL; findingSTEntry = findingSTEntry->getNext())
			{
				if(!strcmp(findingSTEntry->getLexeme(), lexeme))
				{

#ifdef _DEBUG_
					printf("\n") ;
					printf("find the entry of symbol table\n") ;
					printf("lexeme : %s\n",findingSTEntry->getLexeme()) ;
					printf("token : %d\n",findingSTEntry->getToken()) ;
					printf("hashValue : %d\n",hashValue) ;
#endif _DEBUG_

					return findingSTEntry ;
				}
			}
		}

#ifdef _DEBUG_
		printf("\n") ;
		printf("find the entry of symbol table.\n") ;
		printf("Can't find.\n") ;
		printf("lexeme : %s\n", lexeme) ;
		printf("hashValue : %d\n", hashValue) ;
#endif _DEBUG_

		return NULL ;
	}

	/**
	* Dump the symbol table
	*
	* @return none
	*/
	void dump()
	{
		STEntry *dumpSTEntry ;

		printf("\n") ;
		printf("--------------------\n") ;
		for(int i = 0; i < hashTableSize; i++)
		{
			if(hashHead[i] != NULL)
			{
				for(dumpSTEntry = hashHead[i]; dumpSTEntry != NULL; dumpSTEntry = dumpSTEntry->getNext())
				{
					printf("\n") ;
					printf("lexeme : %s\n", dumpSTEntry->getLexeme()) ;
					printf("token : %d\n", dumpSTEntry->getToken()) ;
					printf("hashValue : %d\n", i) ;
				}
			}
		}
		printf("\n") ;
		printf("--------------------\n") ;
	}
} ;
